/***************************************************************************
                          predicatetable.cpp  -  description
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <iostream.h>
#include <iomanip.h>
#include <string.h>
#include "predicatetable.h"
#include "type.h"

PredicateTable::PredicateTable ():
  size ( 0 ),
  array ( NULL ),
  list ( NULL ),
  filledarray ( false ) {
}

PredicateTable::~PredicateTable () {
  if ( filledarray ) {
    for ( int i = 0; i < size; i++ ) {
      delete[] array[i]->name;
      delete[] array[i]->params;
    }
    delete[] array;
  }
  else {
    PredicateDataLE *walk = list, *temp;

    while ( walk ) {
      temp = walk;
      walk = walk->next;
      delete[] temp->data->params;
      delete[] temp->data->name;
      delete temp->data;
      delete temp;
    }
  }
}


/** add a predicate to the table */
void PredicateTable::addPredicate ( char *predicate_name, int size, int *params ) {
  PredicateDataLE *elt = new PredicateDataLE;
  elt->next = list;
  elt->data = new PredicateData;
  elt->data->name = strdup ( predicate_name );
  elt->data->nr = this->size++;
  elt->data->params = params;
  elt->data->size = size;
  elt->data->nmodes = 0;
  elt->data->filln = 0;
  elt->data->key = false;
  elt->data->facttable = NULL;
  predicatenames.add ( predicate_name, elt->data );
  list = elt;
}

PredicateData *PredicateTable::getPredicate ( char *predicate_name ) {
  return (PredicateData*) predicatenames.lookup ( predicate_name );
}

void PredicateTable::fillArray () {
  array = new PredicateData*[size];
  PredicateDataLE *walk = list, *temp;

  for ( int i = 0; i < size; i++ ) {
    array[walk->data->nr] = walk->data;
    temp = walk;
    walk = walk->next;
    delete temp;
  }

  filledarray = true;
}

void PredicateTable::print () {
  int i, j;

  if ( filledarray ) {
    cout << "+--------------+----+------------------+" << endl
         << "| Predicate    | Nr | Param Types      |" << endl
         << "+--------------+----+------------------+" << endl;
    for ( i = 0; i < size; i++ ) {
      cout << "| " << setw ( 12 ) << array[i]->name << " |" << setw ( 3 ) << array[i]->nr << " |";
      for ( j = 0; j < array[i]->size - 1; j++ )
        cout << setw ( 3 ) << array[i]->params[j] << ",";
      cout << setw ( 3 ) << array[i]->params[j]
           << setw ( 16 - j * 4 ) << "|" << endl;
    }
    cout << "+--------------+----+------------------+" << endl;
  }
  else
    cout << "Error: display not possible for unfilled data." << endl;
}

void setFact ( PredicateData *pd, INTTYPE *coors ) {
  for ( int k = 0; k < pd->nmodes; k++ )
    setFact ( pd->modes[k], coors );
}
