/***************************************************************************
                          tree.cpp  -  description
                             -------------------
    begin                : Sat Mar 25 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "tree_global.h"
#include "type.h"

ChildList *root;
int *variables;
int nvar;
int treedepth;
int *typemaxvar;
int **varval; // used to store equivalences and variable values during count
bool **constused; // true in this table if variable is bound to this constant
int keyvarcount = 0;

Node **nodehistory;

void initTree () {
  int i, j;

  Node *rootnode;
  nvar = typetable.size;
  variables = new int[nvar];
  typemaxvar = new int[nvar];
  varval = new int*[nvar];
  for ( i = 0; i < nvar; i++ ) {
    typemaxvar[i] = 0;
    variables[i] = 0;
  }

  root = new ChildList;
  root->next = NULL;
  root->nodes = new Node[1];
  root->size = 1;
  rootnode = root->nodes;
  rootnode->children = NULL;
  rootnode->mode = key->nr;
  SET_SUPPORT ( rootnode, minsup + 1 );
  rootnode->params = new INTTYPE[key->size];
  rootnode->introd = 0;
  SET_NO_EQUIVALENT_NODE ( rootnode );

  for ( i = 0; i < key->size; i++ )
    rootnode->params[i] = typemaxvar[modetable.array[rootnode->mode].params[i].number]++;

  for ( i = 0; i < nvar; i++ )
    varval[i] = new int[typemaxvar[i]];

  treedepth = 1;

    // table filled in and cleaned up by treecount.
  constused = new bool*[nvar];
  for ( i = 0; i < nvar; i++ ) {
    constused[i] = new bool[typetable.array[i]->constnumbers + 1];
    for ( j = 0; j < typetable.array[i]->constnumbers + 1; j++ )
      constused[i][j] = false;
  }
}

void check_equaling ( Node *node ) {
  int predicate = modetable.array[MODE ( node )].predicate;
  ModeData *md;
  Node *nh;
  INTTYPE *n1, *n2;

  for ( int i = 1, k; i < treedepth; i++ ) {
    nh = nodehistory[i];
    md = modetable.array + MODE ( nh );
    if ( md->predicate == predicate ) {
      if ( md->outputs ) {
        for ( k = 0, n1 = node->params, n2 = nh->params;
              k < md->size;
              k++, n1++, n2++ )
          if ( *n1 != *n2 )
            break;
        // identical?
        if ( k == md->size ) {
          DISABLE_NODE ( node );
          return;
        }
      }
    }
  }
}

// disable needed for equivalence search
ChildList *dependentNodes ( int *varsinnode, ModeData *newnodemd, int depth, bool disable ) {
  int parsize = newnodemd->size, varsize = newnodemd->nrinput;
  int vars[parsize], i, j, k;
  int nodessize = 0, prod, varparnr = 0;
  int nrofvar[varsize], mustnrvar[varsize];
  int outputvar[typetable.size];
  int number, newnodei;
  int *outputvarpos;
  ChildList *cl;
  Node *nodes;
  ModeParamData *mpd;

  for ( i = 0; i < typetable.size; i++ )
    outputvar[i] = 0;

    // possibilities for each parameter: constant, variable?
  for ( i = 0, mpd = newnodemd->params; i < parsize; i++, mpd++ ) {
    number = mpd->number;
    switch ( mpd->type ) {
      case Constant: {
        vars[i] = number;
        break;
      }
      case Min: case MinEx: {
        vars[i] = variables[number] + varsinnode[number] + ( outputvar[number]++ );
        break;
      }
      case Plus: {
        int number = newnodemd->params[i].number;
        nrofvar[varparnr] = variables[number] + varsinnode[number] - 1;
        mustnrvar[varparnr] = varsinnode[number];
        varparnr++;
      }
    }
  }
    // varparnr == md->outputnr

    // compute the number of child nodes for this mode decl
  for ( i = 0; i < varparnr; i++ ) {
    prod = 1;
    for ( j = 0; j < i; j++ )
      prod *= nrofvar[j] + 1 - mustnrvar[j];
    prod *= mustnrvar[j];
    for ( j++; j < varparnr; j++ )
        prod *= nrofvar[j] + 1;
    nodessize += prod;
  }

    // create the nodes
  if ( nodessize > 0 ) {
    cl = new ChildList;
    cl->nodes = nodes = new Node[nodessize];
    cl->size = nodessize;
    cl->next = NULL;
    newnodei = 0;
    outputvarpos = newnodemd->inputvarpos;

    vars[outputvarpos[0]] = nrofvar[0] - mustnrvar[0] + 1;
    for ( i = 0; i < varparnr; i++ ) {
      j = 0;

      while ( true )
        if ( j < i )
          if ( vars[outputvarpos[j]] <= nrofvar[j] - mustnrvar[j] ) {
            j++; // goto next variable
            if ( j == i )
              vars[outputvarpos[j]] = nrofvar[j] - mustnrvar[j] + 1;
            else
              vars[outputvarpos[j]] = 0;
          }
          else {
            j--;
	    if ( j < 0 )
	      break;
            vars[outputvarpos[j]]++;
          }
        else
          if ( vars[outputvarpos[j]] <= nrofvar[j] ) {
            j++; // goto next variable
            if ( j == varparnr ) {
              Node *node = nodes + newnodei;
	      ModeParamData *mpd = modetable.array[newnodemd->nr].params;
              node->children = NULL;
              node->disablecount = 0;
              SET_SUPPORT ( node, 0 );
              node->mode = newnodemd->nr;
              node->equin = EQUINNONE;
	      if ( disable )
	        DISABLE_NODE ( node );
              node->params = new INTTYPE[parsize];
              node->introd = depth;
              INTTYPE *np = node->params;
              for ( k = 0; k < parsize; k++, np++, mpd++ ) {
                *np = vars[k];
		if ( mpd->type == Min || mpd->type == MinEx ) {
		  if ( typemaxvar[mpd->number] <= vars[k] )
		    typemaxvar[mpd->number] = vars[k] + 1;
		}
	      }
              newnodei++;

              if ( newnodemd->equalingmode )
                check_equaling ( node );

              j--;
	      if ( j < 0 )
	        break;
              vars[outputvarpos[j]]++;
            }
            else
              vars[outputvarpos[j]] = 0;
          }
          else {
            j--;
	    if ( j < 0 )
	      break;
            vars[outputvarpos[j]]++;
          }
      vars[outputvarpos[0]] = 0;
    }

    return cl;
  }
  else
    return NULL;
}

bool is_equivalent ( Node *oldnode, Node *newnode ) {
  if ( oldnode->mode != newnode->mode )
    return false;
  ModeData *md = modetable.array + newnode->mode;
  int nri = md->nrinput, i, *mdiv = md->inputvarpos;
  ModeParamData *mpd = md->params;
  INTTYPE *newnp = newnode->params, *oldnp = oldnode->params;

  for ( i = 0; i < nri; i++, mdiv++ ) {
    if ( varval[mpd[*mdiv].number][oldnp[*mdiv]] != newnp[*mdiv] )
      return false;
  }

  return true;
}

#define NEXT_CHILD(l,p) \
  {\
    p++;\
    if ( p == l->size ) {\
      l = l->next;\
      p = 0;\
      if ( !l )\
	break;\
    }\
  }


void scan_for_equivalent ( Node *node, ChildList *newl ) {
  int od = node->equin + 1;
  Node *oldnode = nodehistory[node->equin],
       *bordernode = nodehistory[od];
  ChildList *oldl = oldnode->children, *walkl;
  ModeData *md = modetable.array + node->mode;
  ModeParamData *mpd;
  int i, oldp, newp, walkp, prevmode = -1;
  INTTYPE *oldnp, *newnp;
  oldnp = oldnode->params;
  newnp = node->params;
  Node *olnode, *nlnode;
  bool succeed;

  for ( i = 0, mpd = md->params; i < md->size; i++, mpd++, oldnp++, newnp++ )
    if ( mpd->type != Constant )
      varval[mpd->number][*oldnp] = *newnp;

  oldp = 0;
  newp = 0;
  while ( true ) {
     // search for equivalent node for the current node
    olnode = oldl->nodes + oldp;

    if ( olnode->introd == od ) {
      if ( olnode->mode != prevmode ) {
        while ( olnode->mode != newl->nodes[newp].mode )
	  NEXT_CHILD ( newl, newp );
      }
      walkl = newl;
      walkp = newp;
      nlnode = walkl->nodes + walkp;
      while ( !( succeed = is_equivalent ( olnode, nlnode ) ) ) {
	NEXT_CHILD ( walkl, walkp );
        nlnode = walkl->nodes + walkp;
      }
      if ( !succeed )
        cerr << "Algorithm error!!! Please report." << endl;
      if ( olnode == bordernode ) {
        if ( !DISABLED_NODE ( nlnode ) )
          nlnode->equin = node->equin + 1;
        break;
      }
      else
        DISABLE_NODE ( nlnode );
      prevmode = olnode->mode;
    }
    NEXT_CHILD ( oldl, oldp );
  }

  oldnp = oldnode->params;
  newnp = node->params;

  for ( i = 0, mpd = md->params; i < md->size; i++, mpd++, oldnp++, newnp++ )
    if ( mpd->type != Constant )
      varval[mpd->number][*oldnp] = *oldnp;

}

ChildList *dependentNodes ( Node *node, ChildList *&last, int depth ) {
  ModeData *md = modetable.array + MODE ( node ), *md2;
  ModeParamData *mpd = md->params;
  int varsinnode[nvar], i, mdsize = md->size,
      mddepmodsize = md->depmodsize,
      *depmode;
  ChildList *cl, *bg;
  last = NULL;

    // collect variables introduced by this node
  for ( i = 0; i < nvar; i++ )
    varsinnode[i] = 0;
  for ( i = 0; i < mdsize; i++, mpd++ )
    if ( mpd->type == Min || mpd->type == MinEx )
      varsinnode[mpd->number]++;

  if ( mddepmodsize > 0 ) {
    bg = NULL;
    depmode = md->dependentmodes;
    for ( i = 0; i < mddepmodsize; i++, depmode++ ) {
      md2 = modetable.array + *depmode;
      cl = dependentNodes ( varsinnode, md2, depth, md2->number == 0 );
      if ( cl ) {
  	if ( bg )
          last->next = cl;
  	else
          bg = cl;
        last = cl;
      }
    }
  }
  else
    bg = NULL;

  if ( bg && EXISTS_EQUIVALENT_NODE ( node ) )
    scan_for_equivalent ( node, bg );
  
  return bg;
}

void refillRightBrother ( Node *father, Node *n1, Node *n2 ) {
  int newvar[typetable.size], i;
  ModeData *md = modetable.array + MODE ( n1 ),
           *fm = modetable.array + MODE ( father );
  ModeParamData *mpd = md->params,
                *fmpd = fm->params;
  int size = md->size, number,
      fsize = fm->size;
  INTTYPE *np;
  n1->params = new INTTYPE[size];

  for ( i = 0; i < typetable.size; i++ )
    newvar[i] = 0;
  for ( i = 0; i < fsize; i++, fmpd++ )
    if ( fmpd->type == Min || fmpd->type == MinEx )
      newvar[fmpd->number]++;

  np = n1->params;
  for ( i = 0; i < size; i++, np++, mpd++ )
    if ( mpd->type == Min || mpd->type == MinEx ) {
      number = mpd->number;
      *np = variables[number] + newvar[number];
      if ( *np >= typemaxvar[number] )
        typemaxvar[number] = *np + 1;
      newvar[number]++;
    }
    else
      *np = n2->params[i];
}

ChildList *addRightBrothers ( Node *node, ChildList *cwalkp1, int cwalki1, int depth ) {
  ChildList *cwalkp2 = cwalkp1, *nl;
  int cwalks;
  INTTYPE mode = MODE ( node );
  ModeData *md = modetable.array + mode, *md2;
  bool inclself;
  if ( inclself = ( selfduplication && mode != key->nr && md->outputs ) )
    cwalks = cwalki1; // include the node itself
  else
    cwalks = cwalki1 + 1;
  int cwalki2 = cwalks;
  int nchildsize = 0;
  Node *bn, *n;

   // compute number of children
  while ( cwalkp2 ) {
    for ( ; cwalki2 < cwalkp2->size; cwalki2++ ) {
      bn = cwalkp2->nodes + cwalki2;
      md2 = modetable.array + MODE ( bn );
      if ( !DISABLED_NODE ( bn ) && 
           SUPPORT ( bn ) >= minsup && md2->number > 0 &&
           ( md->idnr == -1 || md->idnr != md2->idnr ) )
        nchildsize++;
    }
    cwalki2 = 0;
    cwalkp2 = cwalkp2->next;
  }

  if ( nchildsize > 0 ) {
    ModeData *md3;
    nl = new ChildList;
    nl->size = nchildsize;
    nl->nodes = new Node[nchildsize];
    nl->next = NULL;
    
    if ( inclself ) {
      bn = cwalkp1->nodes + cwalki1;
      md3 = modetable.array + MODE ( bn );
      // self duplicated node should be the last child
      if ( SUPPORT ( bn ) >= minsup && md3->number > 0 && ( md->idnr == -1 || md->idnr != md3->idnr ) ) {
          // if no outputs, check whether this node isn't equal
	  // to a previous node in the path from root to leaf.
	  // (apart from input/output/etc).
        n = nl->nodes + nchildsize - 1;
	*n = *bn;
        SET_SUPPORT ( n, 0 );
	n->disablecount = 0;
	n->children = NULL;
	n->equin = depth - 1;
        if ( md3->outputs )
          // take new output variables!!!
          refillRightBrother ( node, n, bn );
        n++;
      }
    }

    cwalkp2 = cwalkp1;
    cwalki2 = cwalki1 + 1;
    n = nl->nodes;

    while ( cwalkp2 ) {
      bn = cwalkp2->nodes + cwalki2;
      for ( ; cwalki2 < cwalkp2->size; cwalki2++, bn++ ) {
        md3 = modetable.array + MODE ( bn );
        if ( !DISABLED_NODE ( bn ) && 
	     SUPPORT ( bn ) >= minsup && md3->number > 0 &&
	     ( md->idnr == -1 || md->idnr != md3->idnr ) ) {
            // if no outputs, check whether this node isn't equal
	    // to a previous node in the path from root to leaf.
	    // (apart from input/output/etc).
	  *n = *bn;
          SET_SUPPORT ( n, 0 );
	  n->disablecount = 0;
	  n->children = NULL;
          if ( md3->outputs )
            // take new output variables!!!
            refillRightBrother ( node, n, bn );
          n++;
        }
      }
      cwalki2 = 0;
      cwalkp2 = cwalkp2->next;
    }
  }
  else
    nl = NULL;
  
  return nl;
}

bool expandLeaf ( ChildList *childlist, int depth ) {
  ChildList *cwalkp = childlist, *last;
  int cwalki;
  Node *cwalk;
  bool expanded = false;

  while ( cwalkp ) {
    for ( cwalki = 0, cwalk = cwalkp->nodes;
          cwalki < cwalkp->size;
          cwalki++, cwalk++ ) {
      nodehistory[treedepth - 1] = cwalk;
      if ( SUPPORT ( cwalk ) < minsup )
        DISABLE_NODE ( cwalk );
      else {
        cwalk->disablecount = 0;
        modetable.array[MODE ( cwalk )].number--;
        cwalk->children = dependentNodes ( cwalk, last, depth );
        if ( last ) {
          expanded = true;
	    last->next = addRightBrothers ( cwalk, cwalkp, cwalki, depth );
        }
        else
          if ( cwalk->children = addRightBrothers ( cwalk, cwalkp, cwalki, depth ) )
            expanded = true;
          else
            DISABLE_NODE ( cwalk );
        modetable.array[MODE ( cwalk )].number++;
      }
    }
    cwalkp = cwalkp->next;
  }

  return expanded;
}

bool expandTree ( ChildList *childlist, int depth );

bool expandNode ( Node *node, int depth ) {
  int size = typetable.size, shadvar[size], *varsave, i;
  ModeData *md = modetable.array + MODE ( node );
  bool expand;
  ModeParamData *mpd;
  node->disablecount = 0;
  nodehistory[depth - 2] = node;

  if ( md->outputs ) {
      // use new variable table
    varsave = variables;

    for ( i = 0; i < size; i++ )
      shadvar[i] = variables[i];

    variables = shadvar;

    if ( EXISTS_EQUIVALENT_NODE ( node ) ) {
      int ni = node->introd;
      for ( i = 0, mpd = md->params; i < md->size; i++, mpd++ )
        if ( mpd->type == Min || mpd->type == MinEx ) {
          int np = node->params[i];
          varval[mpd->number][np] = nodehistory[ni]->params[i];
          variables[mpd->number] = np + 1;
        }
    }
    else
      for ( i = 0, mpd = md->params; i < md->size; i++, mpd++ )
        if ( mpd->type == Min || mpd->type == MinEx )
          variables[mpd->number] = node->params[i] + 1;

    md->number--;
    expand = expandTree ( node->children, depth );
    md->number++;

    if ( EXISTS_EQUIVALENT_NODE ( node ) ) {
       int *opvp = md->outputvarpos, no = md->nroutput;
       mpd = md->params;
       INTTYPE *np = node->params, opvpn, n;
       for ( i = 0; i < no; i++ ) {
         opvpn = opvp[i];
         n = np[opvpn];
         varval[mpd[opvpn].number][n] = n;
       }
    }

    variables =  varsave;

    return expand;
  }
  else
    return expandTree ( node->children, depth );
}

bool expandTree ( ChildList *childlist, int depth ) {
  if ( depth < treedepth ) {
    bool expand = false;
    ChildList *cwalkp = childlist;
    Node *cwalkn;
    int cwalki;
    while ( cwalkp ) {
      for ( cwalki = 0, cwalkn = cwalkp->nodes;
            cwalki < cwalkp->size;
            cwalki++, cwalkn++ )
        if ( ENABLED_NODE ( cwalkn ) )
          if ( expandNode ( cwalkn, depth + 1 ) )
            expand = true;
          else
            DISABLE_NODE ( cwalkn );
      cwalkp = cwalkp->next;
    }

    return expand;
  }
  else
    return expandLeaf ( childlist, depth );
}

bool expandTree () {
  bool expand;
  int i, j, size = typetable.size;
  nodehistory = new Node*[treedepth];

  for ( i = 0; i < size; i++ )
    for ( j = 0; j < typemaxvar[i]; j++ )
      varval[i][j] = j;

  expand = expandTree ( root, 1 );
/*#ifdef DEBUG
  cout << "+------+---------+ " << endl
       << "| Type | Max Var | " << endl
       << "+------+---------+ " << endl;
  for ( i = 0; i < typetable.size; i++ )
    cout << "|" << setw ( 5 ) << i << " |" << setw ( 8 ) << typemaxvar[i] << " |" << endl;
  cout << "+------+---------+ " << endl;
#endif*/
  treedepth++;

  for ( i = 0; i < size; i++ ) {
    delete[] varval[i];
    varval[i] = new int[typemaxvar[i]];
  }

  delete[] nodehistory;
  return expand;
}

bool countChildren ( ChildList *childlist, int depth );

bool countChild ( Node *node, int depth ) {
  int i, j = 0, typenumber;
  ModeData *md = modetable.array + MODE ( node );
  int nroutput = md->nroutput;
  INTTYPE coors[md->size - nroutput];
  bool prevused[nroutput];
  int cv, *opv;
  INTTYPE *coorlwv;
  ModeParamData *mpd;
  ModeFactCoors *coorlw;

    // fill in variable values and constants
  for ( i = 0, mpd = md->params; i < md->size; i++, mpd++ )
    if ( mpd->type == Constant )
      coors[j++] = mpd->number;
    else
      if ( mpd->type == Plus )
        coors[j++] = varval[mpd->number][node->params[i]];
  
  i = 0;
  if ( nroutput == 0 ) {
    if ( md->table.facttable->getFact ( coors ) &&
         ( !node->children || !countChildren ( node->children, depth ) ) ) {
      node->disablecount = keyvarcount;
      if ( depth == treedepth )
        INC_SUPPORT ( node );
      return false;
    }
    else
      return true;
  }

  coorlw = md->table.modefacttable->getTrueFacts ( coors );
  
  while ( coorlwv = coorlw->value () ) {
    for ( i = 0, opv = md->outputvarpos;
          i < nroutput;
          i++, opv++, coorlwv++ ) {
      mpd = md->params + *opv;
      typenumber = mpd->number;
      cv = *coorlwv;
      if ( mpd->type == MinEx ) {
        if ( constused[typenumber][cv] )
          break;
	prevused[i] = constused[typenumber][cv];
        constused[typenumber][cv] = true;
         // only really bind if \ output, to avoid that there
	 // is a commutativety problem:
	 // in(+A,+B,-C),triangle(+A,\D)
	 // otherwise something different than
	 // triangle(+A,\D),in(+A,+B,-c)
	 // which is confusing when analysing results after the
	 // equivalence mechanism application.
      }
      else
	prevused[i] = constused[typenumber][cv];
      varval[typenumber][node->params[*opv]] = cv;
    }
    if ( i == nroutput ) {
      if ( depth == treedepth /* !node->children */ || !countChildren ( node->children, depth ) ) {
        node->disablecount = keyvarcount;
        if ( depth == treedepth )
          INC_SUPPORT ( node );
	for ( i--, opv--, coorlwv--; i >= 0; i--, opv--, coorlwv-- )
	  constused[md->params[*opv].number][*coorlwv] = prevused[i];
        delete coorlw;
        return false;
      }
    }
    for ( i--, opv--, coorlwv--; i >= 0; i--, opv--, coorlwv-- )
      constused[md->params[*opv].number][*coorlwv] = prevused[i];
  }
  
  delete coorlw;
  return true;
}


// returns true if a node still has to be proved
bool countChildren ( ChildList *childlist, int depth ) {
  ChildList *cwalkp = childlist;
  Node *node;
  int cwalki;
  bool ret = false;

  while ( cwalkp ) {
    for ( cwalki = 0, node = cwalkp->nodes; cwalki < cwalkp->size; cwalki++, node++ )
      if ( ENABLED_NODE ( node ) && node->disablecount < keyvarcount )
        ret = countChild ( node, depth + 1 ) || ret;
    cwalkp = cwalkp->next;
  }
  
  return ret;
}

void countAllRootVars () {
  int i = 0;
  Node *node = root->nodes;
  ModeData *md = modetable.array + MODE ( node );
  int typenumber = md->params[0].number, varnumber = node->params[0];
  keyvarcount = 1;
  
  varval[typenumber][varnumber] = 0;

  int tmpc = 0;
  while ( true ) {
    if ( tmpc % 1000 == 0 ) {
      ERR << tmpc << "\r" << flush;
      ERR_F
    }
    tmpc++;
    if ( varval[typenumber][varnumber] < typetable.array[typenumber]->constnumbers ) {
      if ( i == md->size - 1 ) {
        countChildren ( node->children, 1 );
	varval[typenumber][varnumber]++;
	keyvarcount++;
      }
      else {
        i++;
	typenumber = md->params[i].number;
	varnumber = node->params[i];
	varval[typenumber][varnumber] = 0;
      }
    }
    else {
      i--;
      if ( i < 0 )
        return;
      typenumber = md->params[i].number;
      varnumber = node->params[i];
      varval[typenumber][varnumber]++;
    }
  }
}

void countTree () {
  countAllRootVars ();

  for ( int i = 0; i < typetable.size; i++ )
    for ( int j = 0; j < typemaxvar[i]; j++ )
      varval[i][j] = j;
}

void printTree ( int depth, Node *node, bool infrequent );

void printTree ( int depth, ChildList *children, bool infrequent ) {
  ChildList *walk = children;

  if ( depth / 2 < treedepth )
    while ( walk ) {
      for ( int i = 0; i < walk->size; i++ )
	if ( ENABLED_NODE ( walk->nodes + i ) )
          if ( infrequent || SUPPORT ( walk->nodes + i ) >= minsup )
            printTree ( depth, walk->nodes + i, infrequent );
      walk = walk->next;
    }
}

void printNode ( Node *node ) {
  ModeData *md = modetable.array + MODE ( node );
  PredicateData *pd = predicatetable.array[md->predicate];
  cout << pd->name << "(";
  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case Plus: cout << "+" << pd->params[i] << "." << node->params[i]; break;
      case Min: cout << "-" << pd->params[i] << "." << node->params[i]; break;
      case MinEx: cout << "\\" << pd->params[i] << "." << node->params[i]; break;
      case Constant:
        if ( showconstantnames )
          cout << typetable.getConstantName ( pd->params[i], node->params[i] );
        else
          cout << "#" << pd->params[i] << "." << node->params[i];
        break;
    }

    if ( i < md->size - 1 )
      cout << ", ";
  }
  cout << ")";

}

void printTree ( int depth, Node *node, bool infrequent ) {
  cout << setw ( depth + 2 ) << "+ ";
  printNode ( node );
  cout << "(" << SUPPORT ( node ) / (float) ( keyvarcount - 1 ) << "," << (int) node->equin << ")" << endl;
  printTree ( depth + 2, node->children, infrequent );
}

void printTree ( bool infrequent = true ) {
  printTree ( 0, root, infrequent );
}

bool printMaximalQueries ( Node *node, int depth ) {
  int i;
  ChildList *cl;
  bool print = false;

  nodehistory[depth] = node;
  if ( !node->children ) {
    if ( SUPPORT ( node ) >= minsup )
      print = true;
    else
      return false;
  }
  else {
    cl = node->children;
    while ( cl ) {
      for ( i = 0; i < cl->size; i++ )
        print = printMaximalQueries ( cl->nodes + i, depth + 1 ) || print;
      cl = cl->next;
    }
    print = !print;
  }
  if ( print ) {
    for ( i = 0; i < depth; i++ ) {
      printNode ( nodehistory[i] );
      cout << ", ";
    }
    printNode ( nodehistory[i] );
    cout << endl;
  }
  return true;
}

void printMaximalQueries () {
  nodehistory = new Node*[treedepth];
  printMaximalQueries ( root->nodes, 0 );
  delete[] nodehistory;
}
