/***************************************************************************
                         tree_interesting.cpp  -  Visual Farmer
                             -------------------
    begin                : Fri Aug 4 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "tree_global.h"
#include "type.h"

void printInterestingQueries ( Node *node, int depth ) {
  ChildList *cl;
  int i;

  nodehistory[depth] = node;
  if ( INTERESTING_NODE ( node ) && SUPPORT ( node ) >= minsup ) {
    for ( i = 0; i < depth; i++ ) {
      printNode ( nodehistory[i] );
      cout << ", ";
    }
    printNode ( nodehistory[i] );
    cout << " (" << SUPPORT ( node ) / (float) keyvarcount << ")" << endl;
  }

  cl = node->children;
  while ( cl ) {
    for ( i = 0; i < cl->size; i++ )
      printInterestingQueries ( cl->nodes + i, depth + 1 );
    cl = cl->next;
  }
}

void printInterestingQueries () {
  nodehistory = new Node*[treedepth];
  printInterestingQueries ( root->nodes, 0 );
  delete[] nodehistory;
}

bool determineInterestNoUsedOutput ( int depth ) {
  Node *node = nodehistory[depth];
  ModeData *md = modetable.array + MODE ( node );
  int i, *ovp, nr = md->nroutput;
  INTTYPE *pd = node->params;
  ModeParamData *mpd = md->params;

  ovp = md->outputvarpos;
  for ( i = 0; i < nr; i++, ovp++ )
    if ( varval[mpd[*ovp].number][pd[*ovp]] == 1 )
      return false;

  return true;
}

// subsets which are not found, are:
// p(+X,-Y_1),q(***),...,p(+X,-Y_2)
// ==========
//     ^
//     |
// if removed, an order of atoms results (q before p)
// which is illegal.
//
// after removing, one of isomorphic queries might
// result, of which only one should be present in the tree
void determineInterestWithout ( int depth, int maxdepth ) {
  int minleftout[typetable.size], maxleftout[typetable.size];
  Node *node = nodehistory[depth];
  ModeData *md = modetable.array + MODE ( node );
  int i, j, *ovp, nr = md->nroutput;
  INTTYPE *pd = node->params, *pd1, *pd2, mode1, mode2;
  ModeParamData *mpd = md->params;
  int typenr, varnr, ols;
  ChildList *cl;

  for ( i = 0; i < typetable.size; i++ ) {
    minleftout[i] = typemaxvar[i] + 1;
    maxleftout[i] = 0;
  }

  ovp = md->outputvarpos;
  for ( i = 0; i < nr; i++, ovp++ ) {
    typenr = mpd[*ovp].number;
    varnr = pd[*ovp];
    if ( minleftout[typenr] > varnr )
      minleftout[typenr] = varnr;
    if ( maxleftout[typenr] < varnr )
      maxleftout[typenr] = varnr;
  }

   // search from skipped branch

  cl = nodehistory[depth - 1]->children;
  depth++; // use depth counter to save the node which should be
           // compared
  mode1 = MODE ( nodehistory[depth] );
  pd1 = nodehistory[depth]->params;
  md = modetable.array + mode1;
  nr = md->size;
  ols = cl->size;

   // walk through children
  while ( true ) {
    for ( i = 0; i < ols; i++ ) {
      mode2 = cl->nodes[i].mode;
      pd2 = cl->nodes[i].params;

      // compare the nodes

      if ( mode1 == mode2 ) {
        mpd = md->params;
        for ( j = 0; j < nr; j++, mpd++ )
	  if ( mpd->type == Plus ) {
	    typenr = mpd->number;
	    varnr = pd1[j];
	    if ( varnr >= minleftout[typenr] )
	      varnr += minleftout[typenr] - maxleftout[typenr] - 1;
	    if ( varnr != pd2[j] )
	      break;
	  }
	if ( j == nr ) {
          if ( maxdepth == depth ) {
            SET_NOTMAXIMAL_NODE ( cl->nodes + i );
	    // we have found an equivalent node, and were checking
	    // the last node ==> subset found, check supports
	    if ( INTERESTING (
	           SUPPORT ( cl->nodes + i ),
	           SUPPORT ( nodehistory[maxdepth] )
	         )
	       )
	      SET_NOTINTERESTING_NODE ( cl->nodes + i ); // not interesting, a larger query has comparable support
	    return;
	  }

	  // subsetnode found, go to next node for both (if possible)
	  depth++;
	  cl = cl->nodes[i].children;
	  if ( !cl ) // query can not be found
	    return;
	
	  mode1 = MODE ( nodehistory[depth] );
          pd1 = nodehistory[depth]->params;
	  md = modetable.array + mode1;
	  nr = md->size;
	  break;
	}
      }
    }
    if ( i == ols )
      cl = cl->next;
    if ( cl )
      ols = cl->size;
    else
      break; // stop de loop, not neat...
  }
}

void determineInterestingSubsets ( int depth ) {
  int actusup = SUPPORT ( nodehistory[depth] ), i;

  SET_NOTMAXIMAL_NODE ( nodehistory[depth - 1] );
  if ( INTERESTING (
         SUPPORT ( nodehistory[depth - 1] ),
         actusup
       )
     )
    SET_NOTINTERESTING_NODE ( nodehistory[depth - 1] );

    // check when leaving iteratively the 1th, ..., depth-1th node
    // of the history (if possible, because all outputs unused further on)
    // Note: the root will never be removed, as this is the key

  for ( i = 1; i < depth; i++ )
    if ( determineInterestNoUsedOutput ( i ) )
      determineInterestWithout ( i, depth );
}

void determineInteresting ( Node *node, int depth ) {
  int i;
  ChildList *cl;
  ModeData *md = modetable.array + MODE ( node );
  int *ivp, nr = md->nrinput;
  INTTYPE *pd = node->params;
  int prevused[nr];
  int *pos;
  Node *nnode;
  ModeParamData *mpd = md->params;
  nodehistory[depth] = node;

  ivp = md->inputvarpos;
  for ( i = 0; i < nr; i++, ivp++ ) {
    pos = & ( varval[mpd[*ivp].number][pd[*ivp]] );
    prevused[i] = *pos;
    *pos = 1;
  }

  if ( depth > 0 )
    determineInterestingSubsets ( depth );

  cl = node->children;
  while ( cl ) {
    for ( i = 0; i < cl->size; i++ ) {
      nnode = cl->nodes + i;
      if ( SUPPORT ( nnode ) >= minsup )
        determineInteresting ( nnode, depth + 1 );
    }
    cl = cl->next;
  }

  ivp = md->inputvarpos;
  for ( i = 0; i < nr; i++, ivp++ )
    varval[mpd[*ivp].number][pd[*ivp]] = prevused[i];
}

void determineInteresting () {
  for ( int i = 0; i < typetable.size; i++ )
    for ( int j = 0; j < typemaxvar[i]; j++ )
      varval[i][j] = 0; // not used; 1 is used.
  nodehistory = new Node*[treedepth];
  determineInteresting ( root->nodes, 0 );
  delete[] nodehistory;
}
