/***************************************************************************
                          typetable.cc  -  description
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "output.h"
#include "typetable.h"

TypeTable::TypeTable ():
  actunroftypes ( 0 ),
  filledarray ( false ),
  types ( NULL ),
  number_imm ( true ) {
}

TypeTable::~TypeTable () {
  if ( filledarray )
    delete[] array;
}

void TypeTable::addConstant ( char *constant_name, char *type_name ) {
  TypeData *td = (TypeData*) typenames.lookup ( type_name );
  
  if ( !td ) {
    ERR << "Error: wrong type name (internal)" << endl;
    ERR_F
    return;
  }

  if ( number_imm && constant_name[0] >= '0' && constant_name[0] <= '9' ) {
    ERR << "Error: can not use number as constant name" << endl;
    ERR_F
    return;
  }

  if ( constant_name[0] == '$' ) {
    // float
    float f;
    
    if ( !td->constanttable.table ) {
      td->constanttable.floattable = new FloatTable ( td->maxsize );
      td->names = false;
    }
    else
      if ( td->names ) {
        ERR << "Error: can not add constant name in float type" << endl;
        ERR_F
	return;
      }

    sscanf ( constant_name + 1, "%f", &f );
    td->constanttable.floattable->add ( f, NULL );
  }
  else {
    if ( !td->constanttable.table ) {
      td->constanttable.constantnametable = new NameTable;
      td->names = true;
    }
    else
      if ( !td->names ) {
        ERR << "Error: can not add float to constant names" << endl;
        ERR_F
	return;
      }
    td->constanttable.constantnametable->add ( constant_name, NULL );
  }
  td->constnumbers++;
}

void TypeTable::addType ( char *type_name, int n ) {
  TypeData *td = new TypeData;
  td->constanttable.table = NULL;
  td->constnumbers = 0;
  td->maxsize = n;
  typenames.add ( type_name, td );
  actunroftypes++;
  td->next = types;
  types = td;
}

void TypeTable::getType ( char *type_name, int &typenr, int &nrofconstants ) {
  int ordernumber;
  TypeData *type = (TypeData*) typenames.lookup ( type_name, &ordernumber );
  if ( type ) {
    typenr = ordernumber;
    nrofconstants = type->constnumbers;
  }
  else {
    typenr = -1;
    nrofconstants = -1;
  }
}

// typenr should be filled in
void TypeTable::getConstant ( char *constant_name, int &typenr, int &constantnr ) {
  TypeData *type;

  if ( number_imm && constant_name[0] >= '0' && constant_name[0] <= '9' ) {
    typenr = -1; // unknown, no type... depend on what is later expected
    constantnr = atoi ( constant_name );
  }
  else {
    typenames.lookup ( typenr, (void**) &type );
    if ( constant_name[0] == '$' )
      // float
      
      if ( type->names ) {
        typenr = -1;
	constantnr = -1;
      }
      else
        constantnr = type->constanttable.floattable->ordernumber ( atof ( constant_name + 1 ) );
    else
      constantnr = type->constanttable.constantnametable->ordernumber ( constant_name );
  }
}

// Copyies from a list to an array, AND
// fills in constnumbers if that value is zero.
void TypeTable::fillArray () {
  size = actunroftypes;
  if ( types ) {
    array = new TypeData*[size];
    TypeData *walk = types;
    for ( int i = 0; i < size; i++ ) {
      array[size - i - 1] = walk;
      if ( walk->constnumbers == 0 )
        walk->constnumbers = walk->maxsize;
      walk = walk->next;
    }
  }

  filledarray = true;
}

double TypeTable::getConstantDouble ( int typenr, int constantnr ) {
  TypeData *type = array[typenr];

  return type->constanttable.floattable->lookup ( constantnr );
}


char *TypeTable::getConstantName ( int typenr, int constantnr ) {
  static char buffer[10];

  TypeData *type = array[typenr];

  if ( !type->constanttable.table ) {
    // no further info specified!
    sprintf ( buffer, "%i", constantnr );
    // no name, print number...

    return buffer;
  }
  if ( type->names ) {
    return type->constanttable.constantnametable->lookup ( constantnr );
  }
  else {
    // convert from float to string
    sprintf ( buffer, "$%f", type->constanttable.floattable->lookup ( constantnr ) );

    return buffer;
  }
}

char *TypeTable::getTypeName ( int typenr ) { 
  return typenames.lookup ( typenr );
}

void TypeTable::print () {
  if ( filledarray ) {
    cout << "+------+-----------------+" << endl
         << "| Type | Nr of constants |" << endl
         << "+------+-----------------+" << endl;
    for ( int i = 0; i < size; i++ )
      cout << "| " << setw ( 4 ) << i << " |" << setw( 16 ) << array[i] << " |" << endl;
    cout << "+------+-----------------+" << endl;
  }
  else
    cout << "Error: display not possible for unfilled data." << endl;
}



