/***************************************************************************
                          modefacttable.h  -  description
                             -------------------
    begin                : Tue Apr 11 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MODEFACTTABLE_H
#define MODEFACTTABLE_H
#include "modetable.h"
#include "type.h"

// the ModeFactCoors class is a wrap around for the results of a query.
// by making it a virtual class, several implementations are possible,
// one which stores all the results and allows retreiving them by 
// using nextValue, or another which does only compute the next result
// when nextValue asks for it.
class ModeFactCoors {
  public:
    ModeFactCoors () { }
    virtual ~ModeFactCoors () { }
    virtual INTTYPE *value () = 0;
      // computes/returns the current value, and goes to/prepares for
      // the next value
    virtual bool empty () = 0;
      // does the query have any results left?
};

struct ListModeFactCoorsElement {
  INTTYPE *value;
  ListModeFactCoorsElement *next;
};

// this class implements the easiest case: a class which is a wrap
// around for a list of results, it is used by the default mechanism
// for storing tables, but also other implementations might use it.
class ListModeFactCoors:public ModeFactCoors {
  private:
    ListModeFactCoorsElement *current;
  public:
    ListModeFactCoors ( ListModeFactCoorsElement *first ) {
      current = first;
    }
    INTTYPE *value () {
      if ( current ) {
        INTTYPE *val = current->value;
	current = current->next;
	return val;
      }
      return NULL;
    }
    bool empty () {
      return current;
    }
};

#include "extern.h"

struct TypeTable;

class BaseModeFactTable {
  public:
    BaseModeFactTable () { ; }
    virtual ~BaseModeFactTable () { ; }
    virtual ModeFactCoors* getTrueFacts ( INTTYPE *coor ) = 0;
};

union ModeFactTableCell {
  ListModeFactCoorsElement *truecoors;
  ModeFactTableCell *celldata;
};

/**The ModeFactTable allows fast lookup for a given mode to a fact table by saving search results.
  *@author Siegfried Nijssen
  */
class ModeFactTable:public BaseModeFactTable {
  public:
    ModeFactTable ( TypeTable &typetable, ModeData *modedata  );
    virtual ~ModeFactTable();
  /** Returns a list containing the coordinates of output vars
at which the inputs specified by coors are true. Computes the contents
if not existing */
    ModeFactCoors* getTrueFacts ( INTTYPE *coor ) {
      ModeFactTableCell cell = cells;
      for ( int i = 0; i < inputdim; i++ )
        cell = cell.celldata[coor[i]];
      return new ListModeFactCoors ( cell.truecoors );
    }
     // combined coordinates and value
    void setFact ( INTTYPE *coor );
  private:
    void initTable ( ModeFactTableCell &cells, int dimensions, int *sizes );
    void removeTable ( ModeFactTableCell cells, int dimensions, int *sizes );
    ModeData *modedata;
    ModeFactTableCell cells;
    int inputdim, outputdim;
    int *inputsizes;
};

typedef ModeFactCoors* (*TgetTrueFacts) ( INTTYPE *coor );
TgetTrueFacts extern_lookup_modefacttable ( char *name );

class ExternalModeFactTable:public BaseModeFactTable {
  public:
    ExternalModeFactTable ( char *functionname ) {
      function = extern_lookup_modefacttable ( functionname );
    }
    ModeFactCoors* getTrueFacts ( INTTYPE *coor ) {
      return function ( coor );
    }
  private:
    TgetTrueFacts function;  
};

#endif
