/***************************************************************************
                            -  description
                             -------------------
    begin                : Wed Mar 22 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef MODETABLE_H
#define MODETABLE_H
#include "typetable.h"
#include "facttable.h"
#include "type.h"

class BaseModeFactTable;
class PredicateTable;

enum ModeParamType { Plus, Min, MinEx, Constant };

#define IS_OUTPUT_PARAM(param) ( (param)->type == MinEx || (param)->type == Min )
#define IS_INPUT_PARAM(param) ( (param)->type == Plus || (param)->type == Constant )

struct ModeParamData {
  ModeParamType type;
  int number;
};

struct ModeData {
  int nr;                 // mode declaration number
  int number;             // number of times it can be added
  int predicate;          // predicate number
  ModeParamData *params;  // parameters
  int size;               // nr of parameters
  int *dependentmodes;    // dependent modes
  int depmodsize;         // nr of dependent modes
  int *inputvarpos,       // input variable positions, for fast walking
      *outputvarpos;
  int nrinput, nroutput;  // nr of input variables
  bool outputs;           // has outputs?
  bool equalingmode;      // if no, is there a mode for this predicate
                          // which does at at least one position where
                          // this mode has an input?
  union {
    BaseFactTable *facttable; // if no outputs, binary matrix
    BaseModeFactTable *modefacttable;
                          // otherwise, a matrix containing true output values
  } table;
  bool reuses;            // is another table reused?
  int idnr;
  char *functionname;
};

void setFact ( ModeData *md, INTTYPE *coors );

class ModeTable {
  public:
    ModeTable ();
    ~ModeTable ();
      // stores in the predicate table the number of modes for the
      // predicate
    void fillArray ( PredicateTable &predicatetable );
    ModeData* addMode ( int number, int predicate, int parsize, ModeParamData *params, int idnr, char *functionname );
    void print ();
    ModeData *array;
    int size;
    void fillDependencies ( PredicateTable &predicatetable, TypeTable &typetable );
  private:
    struct ModeDataLE {
      ModeData *data;
      ModeDataLE *next;
    } *list;
    bool filledarray;
};

#endif
