/***************************************************************************
                          tree.h  -  description
                             -------------------
    begin                : Sat Mar 25 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef TREE_H
#define TREE_H
#include "type.h"

struct Node;

struct ChildList {
  Node *nodes;
  short int size;
  ChildList *next;
};

struct Node {
  INTTYPE mode;
  INTTYPE *params;
  unsigned int support;
    // support = '-1' if the node is not interesting
    // array of params. type can be found using mode declaration lookup
  ChildList *children;
  unsigned int disablecount;
  char equin; // number used in the equivalence checking mechanism,
              // denotes the level at which an equivalent node
              // is found.
  char introd; // depth at which this node was introduced, used by
               // the equivalence mechanism
};

#define MODE(node) ((node)->mode)
#define EQUINNONE char(254)
#define EQUINDISABLED char(255)
#define ENABLED_NODE(node) ( (node)->equin != EQUINDISABLED )
#define DISABLED_NODE(node) ( (node)->equin == EQUINDISABLED )
#define DISABLE_NODE(node) (node)->equin = EQUINDISABLED
#define SET_NO_EQUIVALENT_NODE(node) (node)->equin = EQUINNONE
#define EXISTS_EQUIVALENT_NODE(node) ( (node)->equin < EQUINNONE )

#define SUPPORTFLAGS 0x3FFFFFFF
#define NOTINTERESTINGFLAG 0x40000000
#define NOTMAXIMALFLAG 0x80000000
#define SUPPORT(node) ((node)->support & SUPPORTFLAGS)
#define INTERESTING_NODE(node) (!(bool) ( (node)->support & NOTINTERESTINGFLAG ))
#define MAXIMAL_NODE(node) (!(bool) ( (node)->support & NOTMAXIMALFLAG ))
#define SET_SUPPORT(node,supp) (node)->support = (supp)
#define INC_SUPPORT(node) (node)->support++
#define SET_NOTINTERESTING_NODE(node) (node)->support |= NOTINTERESTINGFLAG
#define SET_NOTMAXIMAL_NODE(node) (node)->support |= NOTMAXIMALFLAG

extern ChildList *root;
extern int treedepth;

void initTree ();
void determineInteresting ();
void printTree ( bool infrequent = true );
void printMaximalQueries ();
void printInterestingQueries ();
bool expandTree ();
void countTree ();

typedef void (*subsetCallback) ( int n, bool *added, Node *node );
// find all (legal) subsets of the given sets of nodes, and
// call cb for each of them
void findSubsets ( int n, Node **nodes, subsetCallback cb );

#endif
