/***************************************************************************
                         tree_findsubset.cpp  -  Visual Farmer
                             -------------------
    begin                : Fri Aug 4 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "tree_global.h"
#include "type.h"

Node **newnodehistory;
int *newnewvar;
bool *bp;
int bpn;

bool findSubsets_check_renumber ( Node *node ) {
  ModeData *md = modetable.array + MODE ( node );
  PredicateData *pd = predicatetable.array[md->predicate];
  ModeParamData *mpd = md->params;
  int *ppd = pd->params;
  INTTYPE *np = node->params;

  for ( int i = 0; i < md->size; i++, mpd++, ppd++, np++ ) {
    if ( mpd->type == Plus ) {
      if ( varval[*ppd][*np] == -1 )
         // variable used which is not known in the new query
        return false;
    }
    else
      if ( mpd->type != Constant ) { // min in fact
        varval[*ppd][*np] = newnewvar[*ppd];
        newnewvar[*ppd]++;
      }
  }

  return true;
}

void findSubsets_renumber_undo ( Node *node ) {
  ModeData *md = modetable.array + MODE ( node );
  PredicateData *pd = predicatetable.array[md->predicate];
  ModeParamData *mpd = md->params;
  int *ppd = pd->params;
  INTTYPE *np = node->params;

  for ( int i = 0; i < md->size; i++, mpd++, ppd++, np++ ) {
    if ( mpd->type == Plus ) {
      if ( varval[*ppd][*np] == -1 )
         // variable used which is not known in the new query, stop undo
        return;
    }
    else
      if ( mpd->type != Constant ) {
        varval[*ppd][*np] = -1;
        newnewvar[*ppd]--;
      }
  }
}

bool findSubsets_compare ( Node *treenode, Node *subnode ) {
  ModeData *md = modetable.array + MODE ( subnode );
  PredicateData *pd = predicatetable.array[md->predicate];
  ModeParamData *mpd = md->params;
  int *ppd = pd->params;
  INTTYPE *np = subnode->params;

  if ( MODE ( treenode ) != MODE ( subnode ) )
    return false;

  for ( int i = 0; i < md->size; i++, mpd++, ppd++, np++ ) {
    if ( mpd->type != Constant )
      if ( varval[*ppd][*np] != treenode->params[i] )
        return false;
  }

  return true;
}

Node *findSubsets_find ( Node *parent, Node *node ) {
  ChildList *cl = parent->children;

  while ( cl ) {
    for ( int i = 0; i < cl->size; i++ ) {
      if ( findSubsets_compare ( cl->nodes + i, node ) )
        return cl->nodes + i;
    }
    cl = cl->next;
  }

  return NULL;
}

void findSubsets ( int n, Node **nodes, subsetCallback cb, int ns, bool *added ) {
  Node *nnode;

  if ( n == 0 ) {
    if ( ns ) {
      bool *nadded = new bool[bpn];
      for ( int i = 0; i < bpn; i++ )
        nadded[i] = bp[i];
      cb ( bpn, nadded, newnodehistory[ns - 1] );
    }
  }
  else {
    Node *fn;
    // for the next node, try to add it and try to skip it
    nnode = nodes[0];
    // try to add

    added[0] = true;
    // pass through arguments of nnode, to
    // - determine whether this node can be added
    // - fill in the renumbering of variables
    if ( findSubsets_check_renumber ( nnode ) ) {
      // search the node in the list of children of the given node
      // exception: searching for the root is not possible, take
      // the node itself
      if ( ( ns == 0 && ( fn = nnode ) ) ||
             ( fn = findSubsets_find ( newnodehistory[ns - 1], nnode ) )
         ) {
        newnodehistory[ns] = fn;
        findSubsets ( n - 1, nodes + 1, cb, ns + 1, added + 1 );
      }
    }
    // undo renumbering
    findSubsets_renumber_undo ( nnode );

    added[0] = false;
    // try to skip
    findSubsets ( n - 1, nodes + 1, cb, ns, added + 1 );
  }
}

void findSubsets ( int n, Node **nodes, subsetCallback cb ) {
  int i;
  newnodehistory = new Node*[n];
  newnewvar = new int[typetable.size];
  bp = new bool[n];
  bpn = n;
  for ( i = 0; i < n; i++ )
    bp[i] = false;
  for ( i = 0; i < typetable.size; i++ ) {
    newnewvar[i] = 0;
    for ( int j = 0; j < typemaxvar[i]; j++ )
      varval[i][j] = -1; // variable not renumbered (--> can not be in query)
  }
  findSubsets ( n, nodes, cb, 0, bp );
  delete[] newnodehistory;
  delete[] newnewvar;
  delete[] bp;
}