/***************************************************************************
                         qgraphatom.cpp  -  Visual Farmer
                             -------------------
    begin                : Wed Jul 19 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "qgraphatom.h"
#include <qfontmetrics.h>
#include <qpainter.h>
#include "farmer/globaldata.h"

QGraphAtom::QGraphAtom(int n, Node **nodes,QWidget *parent, const char *name ) : QWidget(parent,name) {
  this->n = n;
  this->nodes = nodes;
  highlight = false;
  setBackgroundColor ( white );
  fillCoors ();
  resize ( w, h );
}

QGraphAtom::~QGraphAtom() {
  delete[] x;
}

#define DISTANCE 5

void QGraphAtom::fillCoors () {
  QFontMetrics qfm = fontMetrics ();
  Node *node = nodes[n];
  int mode = node->mode;
  ModeData *md = modetable.array + mode;
  int predicate = md->predicate;
  PredicateData *pd = predicatetable.array[predicate];
  int tx = DISTANCE;
  QRect iqr = qfm.boundingRect ( pd->name );
  h = iqr.height () + DISTANCE * 3;
  int i;
  char *text;

  x = new int[pd->size + 2];
  x[0] = tx;
  tx += DISTANCE * 2 + iqr.width ();

  for ( i = 0; i < pd->size; i++ ) {
    x[i + 1] = tx;
    if ( md->params[i].type == Constant )
      text = typetable.getConstantName ( pd->params[i], md->params[i].number );
    else
      text = typetable.getTypeName ( pd->params[i] );
    tx += qfm.boundingRect ( text ).width () + DISTANCE * 2;
  }

  w = x[i + 1] = tx - DISTANCE;
}

void QGraphAtom::paintEvent ( QPaintEvent * ) {
  Node *node = nodes[n];
  int mode = node->mode;
  ModeData *md = modetable.array + mode;
  int predicate = md->predicate;
  PredicateData *pd = predicatetable.array[predicate];
  char *text;

  QPainter painter ( this );
  if ( highlight )
    painter.setPen ( QPen ( blue, 2 ) );
  else
    painter.setPen ( QPen ( black, 2 ) );
  painter.drawRect ( 1, 1, w - 1, h - 1 );
  painter.drawText ( x[0], DISTANCE, x[1] - x[0], h, 0, pd->name );

  painter.drawLine ( x[1] - DISTANCE, 0, x[1] - DISTANCE, h );

  if ( highlight )
    painter.setPen ( QPen ( blue, 1 ) );
  else
    painter.setPen ( QPen ( black, 1 ) );

  for ( int i = 0; i < pd->size; i++ ) {
    if ( md->params[i].type == Constant )
      text = typetable.getConstantName ( pd->params[i], md->params[i].number );
    else
      text = typetable.getTypeName ( pd->params[i] );
    painter.drawText ( x[i + 1], DISTANCE, x[i + 2], h, 0, text );
    painter.drawLine ( x[i + 1] - DISTANCE, 0, x[i + 1] - DISTANCE, h );
  }
}

void QGraphAtom::mouseMoveEvent ( QMouseEvent * qme ) {
  QWidget *par = parentWidget ();
  QPainter p ( par );
  p.setRasterOp ( XorROP );
  p.setPen ( QPen ( darkGray, 1 ) );
  p.drawRect ( dragx, dragy, width (), height () );
  dragx = QWidget::x () + qme->x () - tx;
  dragy = QWidget::y () + qme->y () - ty;
  p.drawRect ( dragx, dragy, width (), height () );
}

void QGraphAtom::mousePressEvent ( QMouseEvent *qme ) {
  dragx = QWidget::x ();
  dragy = QWidget::y ();
  tx = qme->x ();
  ty = qme->y ();

  QWidget *par = parentWidget ();
  QPainter p ( par );
  p.setRasterOp ( XorROP );
  p.setPen ( QPen ( darkGray, 1 ) );
  p.drawRect ( dragx, dragy, width (), height () );
}

void QGraphAtom::mouseReleaseEvent ( QMouseEvent *qme ) {
  QWidget *par = parentWidget ();
  QPainter p ( par );
  p.setRasterOp ( XorROP );
  p.setPen ( QPen ( darkGray, 1 ) );
  p.drawRect ( dragx, dragy, width (), height () );

  move ( dragx, dragy );
  par->update ();
}

void QGraphAtom::setHighlight ( bool val ) {
  highlight = val;
  repaint ();
}