/***************************************************************************
                         qgraphquery.cpp  -  Visual Farmer
                             -------------------
    begin                : Wed Jul 19 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "farmer/globaldata.h"
#include "qgraphquery.h"
#define VERTINTERBOXD 75
#define HORZINTERBOXD 75
#define CIRCLESIZE 5

QGraphQuery::QGraphQuery(QWidget *parent, const char *name ) : QScrollView(parent,name) {
  qatoms.setAutoDelete ( true );
  qlines.setAutoDelete ( true );
}

QGraphQuery::~QGraphQuery(){
}

/** Set the query which is shown by the view */
void QGraphQuery::setQuery ( int n, Node **nodes ){
  int atomdepth[n];  // atom --> depth
  int atomstartx[n]; // depth --> startx
  int i;
  int maxstartx = 0;
  QGraphAtom *qga;
  int maxd = 0;

  for ( i = 0; i < n; i++ )
    atomstartx[i] = 0;

  this->n = n;
  this->nodes = nodes;
  qatoms.clear ();
  qlines.clear ();

  for ( i = 0; i < n; i++ )
    addNode ( i, nodes, atomdepth, atomstartx );

  for ( i = 0; i < n; i++ )
    if ( atomstartx[i] > maxstartx )
      maxstartx = atomstartx[i];

  // center the atom displays on the view
  qatoms.first ();
  i = 0;
  while ( qga = qatoms.current () ) {
    int cx = qga->QWidget::x () + ( maxstartx - atomstartx[atomdepth[i]] ) / 2;
    if ( atomdepth[i] > maxd )
      maxd = atomdepth[i];
    qga->move (
      cx,
      qga->QWidget::y ()
    );
    i++;
    qatoms.next ();
  }

  resizeContents ( maxstartx, ( maxd + 1 ) * VERTINTERBOXD ); // test values
  viewport ()->update ();
}

void QGraphQuery::addNode ( int i, Node **nodes, int *atomdepth, int *atomstartx ) {
  QWidget *vp = viewport ();
  QGraphAtom *graphatom = new QGraphAtom ( i, nodes, vp );
  int depth = -1;

  qatoms.append ( graphatom );

  // scan previous nodes for nodes it depends on, to determine
  // the depth and compute arrows.

  Node *node = nodes[i];
  ModeData *md = modetable.array + node->mode;
  PredicateData *pd = predicatetable.array[md->predicate];

  for ( int j = 0; j < pd->size; j++ ) {
    if ( md->params[j].type == Plus ) {
      QGraphAtomArrow *qgaa = new QGraphAtomArrow;
      qgaa->endatom = graphatom;
      qgaa->endx = graphatom->accessX ( j );;
      int d = searchForAtom ( i, pd->params[j], node->params[j], qgaa, atomdepth );
      if ( d > depth )
        depth = d;
      qlines.append ( qgaa );
    }
  }

  atomdepth[i] = depth + 1;
  graphatom->move ( atomstartx[depth + 1], ( depth + 1 ) * VERTINTERBOXD );
  atomstartx[depth + 1] += graphatom->width () + HORZINTERBOXD;

  // temporarily:
  graphatom->show ();
}

// returns the depth of the searched atom
int QGraphQuery::searchForAtom ( int k, int typenr, int varnr, QGraphAtomArrow *graphatomarrow, int *atomdepth ) {
  PredicateData *pd;
  ModeData *md;
  Node *node;

  for ( int i = 0; i < k; i++ ) {
    node = nodes[i];
    md = modetable.array + node->mode;
    pd = predicatetable.array[md->predicate];

    for ( int j = 0; j < pd->size; j++ ) {
      if ( ( md->params[j].type == Min || md->params[j].type == MinEx ) && pd->params[j] == typenr && nodes[i]->params[j] == varnr ) {
        // the searched node is found
        graphatomarrow->beginatom = qatoms.at ( i );
        graphatomarrow->beginx = graphatomarrow->beginatom->accessX ( j );
        return atomdepth[i];
      }
    }
  }

  // AN ERROR occured! An input must be found!

  return -1;
}

void QGraphQuery::drawContentsOffset ( QPainter * p, int offsetx, int offsety, int clipx, int clipy, int clipw, int cliph ) {
  QGraphAtomArrow *qgaa;
  int bx, by, ex, ey;
  p->setPen ( QPen ( lightGray, 2 ) );
  p->setBrush ( lightGray );
  viewport ()->erase ();

  qgaa = qlines.first ();
  while ( qgaa ) {
    bx = qgaa->beginatom->QWidget::x () + qgaa->beginx;
    by = qgaa->beginatom->y () + qgaa->beginatom->height ();
    ex = qgaa->endatom->QWidget::x () + qgaa->endx;
    ey = qgaa->endatom->y ();
    p->drawLine ( bx, by, ex, ey );
    p->drawEllipse ( bx - CIRCLESIZE, by - CIRCLESIZE, CIRCLESIZE * 2, CIRCLESIZE * 2 );
    p->drawEllipse ( ex - CIRCLESIZE, ey - CIRCLESIZE, CIRCLESIZE * 2, CIRCLESIZE * 2 );
    qgaa = qlines.next ();
  }
}

void QGraphQuery::setHighlight ( int n, bool *highlighted ) {
  QGraphAtom *qga = qatoms.first ();
  int i = 0;
  while ( qga ) {
    qga->setHighlight ( highlighted[i] );
    i++;
    qga = qatoms.next ();
  }
}