/***************************************************************************
                         visualfarmerview.cpp  -  Visual Farmer
                             -------------------
    begin                : Mon Jul 10 20:04:47 CEST 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "farmer/globaldata.h"
#include "visualfarmerview.h"
#include "nodetoqstring.h"

VisualfarmerView::VisualfarmerView(QWidget *parent, VisualfarmerDoc *doc) : QSplitter(QSplitter::Vertical, parent)
{
  /** connect doc with the view*/
  connect(doc, SIGNAL(documentChanged()), this, SLOT(slotDocumentChanged()));
  graphquery = new QGraphQuery ( this );
  listview = new QListView ( this );
  listview->setAllColumnsShowFocus ( true );
  listview->addColumn ( "Support", 100 );
  listview->addColumn ( "Query" );
  connect(listview,SIGNAL(selectionChanged(QListViewItem*)), this, SLOT(slotSelectionChanged(QListViewItem*)));
}

VisualfarmerView::~VisualfarmerView()
{
}

void VisualfarmerView::listMaximalQueries ( Node *node, int depth ) {
  int i;
  ChildList *cl;

  nodehistory[depth] = node;
  if ( !node->children ) {
    if ( SUPPORT ( node ) >= minsup && MAXIMAL_NODE ( node ) )
      new queryListViewItem ( listview, depth + 1, nodehistory );
  }
  else {
    cl = node->children;
    while ( cl ) {
      for ( i = 0; i < cl->size; i++ )
        listMaximalQueries ( cl->nodes + i, depth + 1 );
      cl = cl->next;
    }
  }
}

void VisualfarmerView::listMaximalQueries () {
  nodehistory = new Node*[treedepth];
  listMaximalQueries ( root->nodes, 0 );
  delete[] nodehistory;
}

void VisualfarmerView::slotDocumentChanged()
{
  listMaximalQueries ();
}

void VisualfarmerView::slotSelectionChanged ( QListViewItem *qlvi ) {
  if ( qlvi->depth () == 0 ) {
    queryListViewItem *qvi = (queryListViewItem*) qlvi;

    graphquery->setQuery ( qvi->n, qvi->query );
  }
  else {
    subsetListViewItem *slvi = (subsetListViewItem*) qlvi;

    graphquery->setHighlight ( slvi->n, slvi->added );
  }
}

VisualfarmerView::queryListViewItem::queryListViewItem ( QListView *lv, int n, Node **query ):
  QListViewItem ( lv ) {
  this->query = new Node*[n];
  for ( int i = 0; i < n; i++ )
    this->query[i] = query[i];
  this->n = n;
}

const char *VisualfarmerView::queryListViewItem::text ( int column ) const {
  static QString tmp;
  if ( column == 0 )
    return tmp.setNum ( SUPPORT ( query[n - 1] ) );
  else
    if ( column == 1 )
      return toQString ( n, query );
    else
      return NULL;
}

const char *VisualfarmerView::queryListViewItem::key ( int column, bool ascending ) const {
  static QString tmp;
  if ( column == 0 )
    return tmp.sprintf( "%08d", SUPPORT ( query[n - 1] ) );
  else
    if ( column == 1 )
      return tmp.sprintf ( "%08d", n );
    else
      return NULL;
}

VisualfarmerView::queryListViewItem::~queryListViewItem () {
 delete[] query;
}

void VisualfarmerView::queryListViewItem::setup () {
  setExpandable ( true );
  QListViewItem::setup ();
}

VisualfarmerView::queryListViewItem *qlvi;

void addSubset ( int n, bool *added, Node *node ) {
  if ( INTERESTING_NODE ( node ) )
    new VisualfarmerView::subsetListViewItem ( qlvi, n, added, SUPPORT ( node ) );
}

void VisualfarmerView::queryListViewItem::setOpen ( bool o ) {
  if ( o && !childCount() ) {
    qlvi = this;
    findSubsets ( n, query, addSubset );
  }
  QListViewItem::setOpen ( o );
}

VisualfarmerView::subsetListViewItem::subsetListViewItem ( queryListViewItem *qlvi, int n, bool *added, int support ):
  QListViewItem ( qlvi ) {
  this->added = added;
  this->n = n;
  this->support = support;
}

const char *VisualfarmerView::subsetListViewItem::text ( int column ) const {
  static QString tmp;
  if ( column == 0 )
    return tmp.setNum ( support );
  else
    if ( column == 1 )
      return toQString ( n, ((queryListViewItem*) parent ())->query, added );
    else
      return NULL;
}

const char *VisualfarmerView::subsetListViewItem::key ( int column, bool ascending ) const {
  static QString tmp;
  if ( column == 0 )
    return tmp.sprintf( "%08d", support );
  else
    if ( column == 1 )
      return tmp.sprintf ( "%08d", n );
    else
      return NULL;
}

VisualfarmerView::subsetListViewItem::~subsetListViewItem () {
 delete[] added;
}
